﻿using System;
using System.Collections.Generic;
using System.Linq;

namespace LinqToCsv
{
    /// <summary>
    /// Exception for CSV parser
    /// </summary>
    public class CsvParseException : Exception
    {
        /// <summary>
        /// Initializes a new instance of the <see cref="CsvParseException"/> class.
        /// </summary>
        /// <param name="message">The message.</param>
        public CsvParseException(string message)
            : base(message)
        {
        }
    }

    /// <summary>
    /// CSV extension for string class
    /// </summary>
    public static class CsvStringExtension
    {
        private enum State
        {
            AtBeginningOfToken,
            InNonQuotedToken,
            InQuotedToken,
            ExpectingComma,
            ExpectingQuote
        };

        /// <summary>
        /// Splits a CSV string (comma separated values, quotes).
        /// </summary>
        /// <param name="source">The source.</param>
        /// <returns></returns>
        public static string[] CsvSplit(this String source)
        {
            var splitString = new List<string>();
            List<int> slashesToRemove = null;
            State state = State.AtBeginningOfToken;
            char[] sourceCharArray = source.ToCharArray();
            int tokenStart = 0;
            int len = sourceCharArray.Length;
            for (int i = 0; i < len; ++i)
            {
                switch (state)
                {
                    case State.AtBeginningOfToken:
                        if (sourceCharArray[i] == '"')
                        {
                            state = State.InQuotedToken;
                            slashesToRemove = new List<int>();
                            continue;
                        }
                        if (sourceCharArray[i] == ',')
                        {
                            splitString.Add("");
                            tokenStart = i + 1;
                            continue;
                        }
                        state = State.InNonQuotedToken;
                        continue;
                    case State.InNonQuotedToken:
                        if (sourceCharArray[i] == ',')
                        {
                            splitString.Add(
                                source.Substring(tokenStart, i - tokenStart));
                            state = State.AtBeginningOfToken;
                            tokenStart = i + 1;
                        }
                        continue;
                    case State.InQuotedToken:
                        if (sourceCharArray[i] == '"')
                        {
                            state = State.ExpectingComma;
                            continue;
                        }
                        if (sourceCharArray[i] == '\\')
                        {
                            state = State.ExpectingQuote;
                            if (slashesToRemove != null) slashesToRemove.Add(i - tokenStart);
                            continue;
                        }
                        continue;
                    case State.ExpectingComma:
                        if (sourceCharArray[i] != ',')
                            throw new CsvParseException("Expecting comma");
                        string stringWithSlashes =
                            source.Substring(tokenStart, i - tokenStart);
                        if (slashesToRemove != null)
                            foreach (int item in slashesToRemove.Reverse<int>())
                                stringWithSlashes =
                                    stringWithSlashes.Remove(item, 1);
                        splitString.Add(
                            stringWithSlashes.Substring(1,
                                stringWithSlashes.Length - 2));
                        state = State.AtBeginningOfToken;
                        tokenStart = i + 1;
                        continue;
                    case State.ExpectingQuote:
                        if (sourceCharArray[i] == '\"')
                        {
                            state = State.InQuotedToken;
                            continue;
                        }
                        throw new CsvParseException("Expecting quote");
                }
            }
            switch (state)
            {
                case State.AtBeginningOfToken:
                    splitString.Add("");
                    return splitString.ToArray();
                case State.InNonQuotedToken:
                    splitString.Add(
                        source.Substring(tokenStart,
                            source.Length - tokenStart));
                    return splitString.ToArray();
                case State.InQuotedToken:
                    throw new CsvParseException("Expecting ending quote");
                case State.ExpectingComma:
                    string stringWithSlashes =
                        source.Substring(tokenStart, source.Length - tokenStart);
                    if (slashesToRemove != null)
                        foreach (int item in slashesToRemove.Reverse<int>())
                            stringWithSlashes = stringWithSlashes.Remove(item, 1);
                    splitString.Add(
                        stringWithSlashes.Substring(1,
                            stringWithSlashes.Length - 2));
                    return splitString.ToArray();
                case State.ExpectingQuote:
                    throw new CsvParseException("Expecting quote");
            }
            throw new CsvParseException("Unexpected error");
        }

    }
}
